--
-- Register addresses in ds_axi space:
--
-- 0x00 IENABLE register - Enabled internal interrupts
--     [31:0]       R/W    Reading returns value of IENABLE register
--                         Writing changes value of IENABLE register
--
-- 0x04 IENABLE_SET
--     [31:0]       R/W1S  Reading returns value of IENABLE register
--                         Writing a 1 at a particular bit position sets that bit in IENABLE register
--
-- 0x08 IENABLE_CLR
--     [31:0]       R/W1C  Reading returns value of IENABLE register
--                         Writing a 1 at a particular bit position clears that bit in IENABLE register
--
-- 0x0C ISTATUS register - Pending internal interrupts
--     [31:0]       R/W1C  Reading returns value of ISTATUS register (i.e. pending interrupts)
--                         Writing a 1 at a particular bit position clears that pending interrupt in ISTATUS
--
-- 0x10 TEST register - Bit 0 is internal interrupt 0
--     [0]          R/WO   Reading returns 0
--                         Writing a 1 generates a pulse on internal interrupt 0
--     [31:1]       R0/MBZ Reserved
--
-- 0x20 PERIOD - Generates a pulse periodically on internal interrupt 1
--     [30:0]       R/W    Pulse generation period minus 1, in milliseconds
--     [31]         R/W    If 1, enables generation of internal interrupt 1 pulse
--
-- Other registers are reserved and must not be written to.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

entity itest is
  port(
    -- MPTL
    mptl_clk_p               : in    std_logic;
    mptl_clk_n               : in    std_logic;
    mptl_b2t_p               : in    std_logic_vector(3 downto 0);
    mptl_b2t_n               : in    std_logic_vector(3 downto 0);
    mptl_t2b_p               : out   std_logic_vector(3 downto 0);
    mptl_t2b_n               : out   std_logic_vector(3 downto 0);
    -- MPTL initialization
    mptl_bridge_gtp_online_l : in    std_logic;
    mptl_target_configured_l : out   std_logic;
    mptl_target_gtp_online_l : out   std_logic;
    -- FPGA to Host interrupt request
    firq_l                   : out   std_logic;
    -- 300 MHz reference clock
    refclk300_p              : in    std_logic;
    refclk300_n              : in    std_logic
  );
end entity;

architecture admxrcku1 of itest is

  component mptl_if
    port(
      -- Reset in
      async_sys_rst            : in    std_logic;    
      -- AXI4
      aclk                     : out   std_logic;
      aresetn                  : out   std_logic;
      -- Direct slave
      ds_axi_awaddr            : out   std_logic_vector(63 downto 0);
      ds_axi_awlen             : out   std_logic_vector(7 downto 0);
      ds_axi_awsize            : out   std_logic_vector(2 downto 0);
      ds_axi_awburst           : out   std_logic_vector(1 downto 0);
      ds_axi_awcache           : out   std_logic_vector(3 downto 0);
      ds_axi_awprot            : out   std_logic_vector(2 downto 0);
      ds_axi_awvalid           : out   std_logic;
      ds_axi_wdata             : out   std_logic_vector(127 downto 0);
      ds_axi_wstrb             : out   std_logic_vector(15 downto 0);
      ds_axi_wlast             : out   std_logic;
      ds_axi_wvalid            : out   std_logic;
      ds_axi_bready            : out   std_logic;
      ds_axi_araddr            : out   std_logic_vector(63 downto 0);
      ds_axi_arlen             : out   std_logic_vector(7 downto 0);
      ds_axi_arsize            : out   std_logic_vector(2 downto 0);
      ds_axi_arburst           : out   std_logic_vector(1 downto 0);
      ds_axi_arcache           : out   std_logic_vector(3 downto 0);
      ds_axi_arprot            : out   std_logic_vector(2 downto 0);
      ds_axi_arvalid           : out   std_logic;
      ds_axi_rready            : out   std_logic;
      ds_axi_awready           : in    std_logic;
      ds_axi_wready            : in    std_logic;
      ds_axi_bresp             : in    std_logic_vector(1 downto 0);
      ds_axi_bvalid            : in    std_logic;
      ds_axi_arready           : in    std_logic;
      ds_axi_rdata             : in    std_logic_vector(127 downto 0);
      ds_axi_rresp             : in    std_logic_vector(1 downto 0);
      ds_axi_rlast             : in    std_logic;
      ds_axi_rvalid            : in    std_logic;
      ds_axi_awlock            : out   std_logic_vector(0 downto 0);
      ds_axi_awqos             : out   std_logic_vector(3 downto 0);
      ds_axi_arlock            : out   std_logic_vector(0 downto 0);
      ds_axi_arqos             : out   std_logic_vector(3 downto 0);
      ds_axi_awregion          : out   std_logic_vector(3 downto 0);
      ds_axi_arregion          : out   std_logic_vector(3 downto 0);
      -- MPTL
      mptl_tx_p                : out   std_logic_vector(3 downto 0);
      mptl_tx_n                : out   std_logic_vector(3 downto 0);
      mptl_rx_p                : in    std_logic_vector(3 downto 0);
      mptl_rx_n                : in    std_logic_vector(3 downto 0);
      mptl_clk_p               : in    std_logic;
      mptl_clk_n               : in    std_logic;
      mptl_clk_out             : out   std_logic;
      -- MPTL initialisation
      mptl_target_configured_n : out   std_logic;
      mptl_target_gtp_online_n : out   std_logic;
      mptl_bridge_gtp_online_n : in    std_logic;
      -- MPTL sideband
      gpio_t2b                 : in    std_logic_vector(15 downto 0);
      gpio_b2t                 : out   std_logic_vector(15 downto 0)
    );
  end component;

  component axi4_128_to_32
    port(
      s_axi_aclk     : in    std_logic;
      s_axi_aresetn  : in    std_logic;
      s_axi_awaddr   : in    std_logic_vector(63 downto 0);
      s_axi_awlen    : in    std_logic_vector(7 downto 0);
      s_axi_awsize   : in    std_logic_vector(2 downto 0);
      s_axi_awburst  : in    std_logic_vector(1 downto 0);
      s_axi_awlock   : in    std_logic_vector(0 downto 0);
      s_axi_awcache  : in    std_logic_vector(3 downto 0);
      s_axi_awprot   : in    std_logic_vector(2 downto 0);
      s_axi_awregion : in    std_logic_vector(3 downto 0);
      s_axi_awqos    : in    std_logic_vector(3 downto 0);
      s_axi_awvalid  : in    std_logic;
      s_axi_awready  : out   std_logic;
      s_axi_wdata    : in    std_logic_vector(127 downto 0);
      s_axi_wstrb    : in    std_logic_vector(15 downto 0);
      s_axi_wlast    : in    std_logic;
      s_axi_wvalid   : in    std_logic;
      s_axi_wready   : out   std_logic;
      s_axi_bresp    : out   std_logic_vector(1 downto 0);
      s_axi_bvalid   : out   std_logic;
      s_axi_bready   : in    std_logic;
      s_axi_araddr   : in    std_logic_vector(63 downto 0);
      s_axi_arlen    : in    std_logic_vector(7 downto 0);
      s_axi_arsize   : in    std_logic_vector(2 downto 0);
      s_axi_arburst  : in    std_logic_vector(1 downto 0);
      s_axi_arlock   : in    std_logic_vector(0 downto 0);
      s_axi_arcache  : in    std_logic_vector(3 downto 0);
      s_axi_arprot   : in    std_logic_vector(2 downto 0);
      s_axi_arregion : in    std_logic_vector(3 downto 0);
      s_axi_arqos    : in    std_logic_vector(3 downto 0);
      s_axi_arvalid  : in    std_logic;
      s_axi_arready  : out   std_logic;
      s_axi_rdata    : out   std_logic_vector(127 downto 0);
      s_axi_rresp    : out   std_logic_vector(1 downto 0);
      s_axi_rlast    : out   std_logic;
      s_axi_rvalid   : out   std_logic;
      s_axi_rready   : in    std_logic;
      m_axi_awaddr   : out   std_logic_vector(63 downto 0);
      m_axi_awlen    : out   std_logic_vector(7 downto 0);
      m_axi_awsize   : out   std_logic_vector(2 downto 0);
      m_axi_awburst  : out   std_logic_vector(1 downto 0);
      m_axi_awlock   : out   std_logic_vector(0 downto 0);
      m_axi_awcache  : out   std_logic_vector(3 downto 0);
      m_axi_awprot   : out   std_logic_vector(2 downto 0);
      m_axi_awregion : out   std_logic_vector(3 downto 0);
      m_axi_awqos    : out   std_logic_vector(3 downto 0);
      m_axi_awvalid  : out   std_logic;
      m_axi_awready  : in    std_logic;
      m_axi_wdata    : out   std_logic_vector(31 downto 0);
      m_axi_wstrb    : out   std_logic_vector(3 downto 0);
      m_axi_wlast    : out   std_logic;
      m_axi_wvalid   : out   std_logic;
      m_axi_wready   : in    std_logic;
      m_axi_bresp    : in    std_logic_vector(1 downto 0);
      m_axi_bvalid   : in    std_logic;
      m_axi_bready   : out   std_logic;
      m_axi_araddr   : out   std_logic_vector(63 downto 0);
      m_axi_arlen    : out   std_logic_vector(7 downto 0);
      m_axi_arsize   : out   std_logic_vector(2 downto 0);
      m_axi_arburst  : out   std_logic_vector(1 downto 0);
      m_axi_arlock   : out   std_logic_vector(0 downto 0);
      m_axi_arcache  : out   std_logic_vector(3 downto 0);
      m_axi_arprot   : out   std_logic_vector(2 downto 0);
      m_axi_arregion : out   std_logic_vector(3 downto 0);
      m_axi_arqos    : out   std_logic_vector(3 downto 0);
      m_axi_arvalid  : out   std_logic;
      m_axi_arready  : in    std_logic;
      m_axi_rdata    : in    std_logic_vector(31 downto 0);
      m_axi_rresp    : in    std_logic_vector(1 downto 0);
      m_axi_rlast    : in    std_logic;
      m_axi_rvalid   : in    std_logic;
      m_axi_rready   : out   std_logic
    );
  end component;

  component axi4_to_axi4l
    port(
      aclk           : in    std_logic;
      aresetn        : in    std_logic;
      s_axi_awaddr   : in    std_logic_vector(63 downto 0);
      s_axi_awlen    : in    std_logic_vector(7 downto 0);
      s_axi_awsize   : in    std_logic_vector(2 downto 0);
      s_axi_awburst  : in    std_logic_vector(1 downto 0);
      s_axi_awlock   : in    std_logic_vector(0 downto 0);
      s_axi_awcache  : in    std_logic_vector(3 downto 0);
      s_axi_awprot   : in    std_logic_vector(2 downto 0);
      s_axi_awregion : in    std_logic_vector(3 downto 0);
      s_axi_awqos    : in    std_logic_vector(3 downto 0);
      s_axi_awvalid  : in    std_logic;
      s_axi_awready  : out   std_logic;
      s_axi_wdata    : in    std_logic_vector(31 downto 0);
      s_axi_wstrb    : in    std_logic_vector(3 downto 0);
      s_axi_wlast    : in    std_logic;
      s_axi_wvalid   : in    std_logic;
      s_axi_wready   : out   std_logic;
      s_axi_bresp    : out   std_logic_vector(1 downto 0);
      s_axi_bvalid   : out   std_logic;
      s_axi_bready   : in    std_logic;
      s_axi_araddr   : in    std_logic_vector(63 downto 0);
      s_axi_arlen    : in    std_logic_vector(7 downto 0);
      s_axi_arsize   : in    std_logic_vector(2 downto 0);
      s_axi_arburst  : in    std_logic_vector(1 downto 0);
      s_axi_arlock   : in    std_logic_vector(0 downto 0);
      s_axi_arcache  : in    std_logic_vector(3 downto 0);
      s_axi_arprot   : in    std_logic_vector(2 downto 0);
      s_axi_arregion : in    std_logic_vector(3 downto 0);
      s_axi_arqos    : in    std_logic_vector(3 downto 0);
      s_axi_arvalid  : in    std_logic;
      s_axi_arready  : out   std_logic;
      s_axi_rdata    : out   std_logic_vector(31 downto 0);
      s_axi_rresp    : out   std_logic_vector(1 downto 0);
      s_axi_rlast    : out   std_logic;
      s_axi_rvalid   : out   std_logic;
      s_axi_rready   : in    std_logic;
      m_axi_awaddr   : out   std_logic_vector(63 downto 0);
      m_axi_awprot   : out   std_logic_vector(2 downto 0);
      m_axi_awvalid  : out   std_logic;
      m_axi_awready  : in    std_logic;
      m_axi_wdata    : out   std_logic_vector(31 downto 0);
      m_axi_wstrb    : out   std_logic_vector(3 downto 0);
      m_axi_wvalid   : out   std_logic;
      m_axi_wready   : in    std_logic;
      m_axi_bresp    : in    std_logic_vector(1 downto 0);
      m_axi_bvalid   : in    std_logic;
      m_axi_bready   : out   std_logic;
      m_axi_araddr   : out   std_logic_vector(63 downto 0);
      m_axi_arprot   : out   std_logic_vector(2 downto 0);
      m_axi_arvalid  : out   std_logic;
      m_axi_arready  : in    std_logic;
      m_axi_rdata    : in    std_logic_vector(31 downto 0);
      m_axi_rresp    : in    std_logic_vector(1 downto 0);
      m_axi_rvalid   : in    std_logic;
      m_axi_rready   : out   std_logic
    );
  end component;

  component axi4l_to_regs
    port (
      s_axi_aclk    : in    std_logic;
      s_axi_aresetn : in    std_logic;
      s_axi_awaddr  : in    std_logic_vector(11 downto 0);
      s_axi_awprot  : in    std_logic_vector(2 downto 0);
      s_axi_awvalid : in    std_logic;
      s_axi_awready : out   std_logic;
      s_axi_wdata   : in    std_logic_vector(31 downto 0);
      s_axi_wstrb   : in    std_logic_vector(3 downto 0);
      s_axi_wvalid  : in    std_logic;
      s_axi_wready  : out   std_logic;
      s_axi_bresp   : out   std_logic_vector(1 downto 0);
      s_axi_bvalid  : out   std_logic;
      s_axi_bready  : in    std_logic;
      s_axi_araddr  : in    std_logic_vector(11 downto 0);
      s_axi_arprot  : in    std_logic_vector(2 downto 0);
      s_axi_arvalid : in    std_logic;
      s_axi_arready : out   std_logic;
      s_axi_rdata   : out   std_logic_vector(31 downto 0);
      s_axi_rresp   : out   std_logic_vector(1 downto 0);
      s_axi_rvalid  : out   std_logic;
      s_axi_rready  : in    std_logic;
      bram_rst_a    : out   std_logic;
      bram_clk_a    : out   std_logic;
      bram_en_a     : out   std_logic;
      bram_we_a     : out   std_logic_vector(3 downto 0);
      bram_addr_a   : out   std_logic_vector(11 downto 0);
      bram_wrdata_a : out   std_logic_vector(31 downto 0);
      bram_rddata_a : in    std_logic_vector(31 downto 0)
    );
  end component;


  -- 300 MHz buffered reference clock
  signal refclk300_ibuf : std_logic;
  signal refclk300 : std_logic;
  
  -- MPTL to AXI interface signals
  signal aclk : std_logic;
  signal aresetn : std_logic;

  signal mptl_gpio_b2t : std_logic_vector(15 downto 0);
  signal mptl_gpio_t2b : std_logic_vector(15 downto 0) := (others => '0');

  signal ds_axi_awaddr : std_logic_vector(63 downto 0);
  signal ds_axi_awlen : std_logic_vector(7 downto 0);
  signal ds_axi_awsize : std_logic_vector(2 downto 0);
  signal ds_axi_awburst : std_logic_vector(1 downto 0);
  signal ds_axi_awcache : std_logic_vector(3 downto 0);
  signal ds_axi_awprot : std_logic_vector(2 downto 0);
  signal ds_axi_awvalid : std_logic;
  signal ds_axi_wdata : std_logic_vector(127 downto 0);
  signal ds_axi_wstrb : std_logic_vector(15 downto 0);
  signal ds_axi_wlast : std_logic;
  signal ds_axi_wvalid : std_logic;
  signal ds_axi_bready : std_logic;
  signal ds_axi_araddr : std_logic_vector(63 downto 0);
  signal ds_axi_arlen : std_logic_vector(7 downto 0);
  signal ds_axi_arsize : std_logic_vector(2 downto 0);
  signal ds_axi_arburst : std_logic_vector(1 downto 0);
  signal ds_axi_arcache : std_logic_vector(3 downto 0);
  signal ds_axi_arprot : std_logic_vector(2 downto 0);
  signal ds_axi_arvalid : std_logic;
  signal ds_axi_rready : std_logic;
  signal ds_axi_awready : std_logic;
  signal ds_axi_wready : std_logic;
  signal ds_axi_bresp : std_logic_vector(1 downto 0);
  signal ds_axi_bvalid : std_logic;
  signal ds_axi_arready : std_logic;
  signal ds_axi_rdata : std_logic_vector(127 downto 0);
  signal ds_axi_rresp : std_logic_vector(1 downto 0);
  signal ds_axi_rlast : std_logic;
  signal ds_axi_rvalid : std_logic;
  signal ds_axi_awlock : std_logic_vector(0 downto 0);
  signal ds_axi_awqos : std_logic_vector(3 downto 0);
  signal ds_axi_arlock : std_logic_vector(0 downto 0);
  signal ds_axi_arqos : std_logic_vector(3 downto 0);
  signal ds_axi_awregion : std_logic_vector(3 downto 0);
  signal ds_axi_arregion : std_logic_vector(3 downto 0);

  -- Width-reduced version of ds_axi_* signals
  signal ds_axi32_awaddr : std_logic_vector(63 downto 0);
  signal ds_axi32_awlen : std_logic_vector(7 downto 0);
  signal ds_axi32_awsize : std_logic_vector(2 downto 0);
  signal ds_axi32_awburst : std_logic_vector(1 downto 0);
  signal ds_axi32_awcache : std_logic_vector(3 downto 0);
  signal ds_axi32_awprot : std_logic_vector(2 downto 0);
  signal ds_axi32_awvalid : std_logic;
  signal ds_axi32_wdata : std_logic_vector(31 downto 0);
  signal ds_axi32_wstrb : std_logic_vector(3 downto 0);
  signal ds_axi32_wlast : std_logic;
  signal ds_axi32_wvalid : std_logic;
  signal ds_axi32_bready : std_logic;
  signal ds_axi32_araddr : std_logic_vector(63 downto 0);
  signal ds_axi32_arlen : std_logic_vector(7 downto 0);
  signal ds_axi32_arsize : std_logic_vector(2 downto 0);
  signal ds_axi32_arburst : std_logic_vector(1 downto 0);
  signal ds_axi32_arcache : std_logic_vector(3 downto 0);
  signal ds_axi32_arprot : std_logic_vector(2 downto 0);
  signal ds_axi32_arvalid : std_logic;
  signal ds_axi32_rready : std_logic;
  signal ds_axi32_awready : std_logic;
  signal ds_axi32_wready : std_logic;
  signal ds_axi32_bresp : std_logic_vector(1 downto 0);
  signal ds_axi32_bvalid : std_logic;
  signal ds_axi32_arready : std_logic;
  signal ds_axi32_rdata : std_logic_vector(31 downto 0);
  signal ds_axi32_rresp : std_logic_vector(1 downto 0);
  signal ds_axi32_rlast : std_logic;
  signal ds_axi32_rvalid : std_logic;
  signal ds_axi32_awlock : std_logic_vector(0 downto 0);
  signal ds_axi32_awqos : std_logic_vector(3 downto 0);
  signal ds_axi32_arlock : std_logic_vector(0 downto 0);
  signal ds_axi32_arqos : std_logic_vector(3 downto 0);
  signal ds_axi32_awregion : std_logic_vector(3 downto 0);
  signal ds_axi32_arregion : std_logic_vector(3 downto 0);

  -- AXI4 Lite version of ds_axi32_* signals
  signal ds_axil_awaddr : std_logic_vector(63 downto 0);
  signal ds_axil_awprot : std_logic_vector(2 downto 0);
  signal ds_axil_awvalid : std_logic;
  signal ds_axil_wdata : std_logic_vector(31 downto 0);
  signal ds_axil_wstrb : std_logic_vector(3 downto 0);
  signal ds_axil_wvalid : std_logic;
  signal ds_axil_bready : std_logic;
  signal ds_axil_araddr : std_logic_vector(63 downto 0);
  signal ds_axil_arprot : std_logic_vector(2 downto 0);
  signal ds_axil_arvalid : std_logic;
  signal ds_axil_rready : std_logic;
  signal ds_axil_awready : std_logic;
  signal ds_axil_wready : std_logic;
  signal ds_axil_bresp : std_logic_vector(1 downto 0);
  signal ds_axil_bvalid : std_logic;
  signal ds_axil_arready : std_logic;
  signal ds_axil_rdata : std_logic_vector(31 downto 0);
  signal ds_axil_rresp : std_logic_vector(1 downto 0);
  signal ds_axil_rvalid : std_logic;

  -- Register read and write signals
  constant REG_DATA_WIDTH : natural := 32;
  constant REG_BE_WIDTH : natural := 4;
  constant REG_BE_WIDTH_LOG2 : natural := 2;
  constant REG_ADDR_WIDTH    : natural range 1 to 22 := 3; -- Byte address width
  signal reg_clk_a : std_logic;
  signal reg_rst_a : std_logic;
  signal reg_rddata_a : std_logic_vector(REG_DATA_WIDTH-1 downto 0);
  signal reg_wrdata_a : std_logic_vector(REG_DATA_WIDTH-1 downto 0);
  signal reg_we_a : std_logic_vector(REG_BE_WIDTH-1 downto 0);
  signal reg_en_a : std_logic;
  signal reg_addr_a : std_logic_vector(11 downto 0); -- Byte address

  -- Registers
  signal reg_ienable : std_logic_vector(REG_DATA_WIDTH-1 downto 0) := (others => '0');
  signal reg_istatus : std_logic_vector(REG_DATA_WIDTH-1 downto 0) := (others => '0');
  signal reg_test : std_logic_vector(REG_DATA_WIDTH-1 downto 0) := (others => '0');
  signal reg_period : std_logic_vector(REG_DATA_WIDTH-1 downto 0) := (others => '0');
  
  -- Rearm pulse for firq_l
  signal rearm_firq : std_logic := '0';
  
  -- Internal interrupt sources
  signal internal_interrupts : std_logic_vector(31 downto 0);
  
  -- Sufficient to count up to 300 000
  signal ms_counter : std_logic_vector(18 downto 0) := (others => '0');
  signal ms_counter_tc : std_logic := '0';
  signal ms_counter_tc_latched : std_logic := '0';

  -- Signals for transferring ms_counter_tc_latched into reg_clk_a clock domain
  signal ms_pulse_xfer0 : std_logic := '0';
  signal ms_pulse_xfer1 : std_logic := '0';
  signal ms_pulse_xfer2 : std_logic := '0';
  signal ms_pulse_xfered : std_logic := '0';
  
  -- Timer interrupt; pulses every period_reg[30:0]+1 milliseconds
  signal timer_count : std_logic_vector(30 downto 0) := (others => '0');
  signal timer_interrupt : std_logic := '0';
  
  signal logic0 : std_logic := '0';
  signal logic1 : std_logic := '1';


begin

  --
  -- The only-and-only MPTL interface instance.
  --

  mptl_if_inst : mptl_if
    port map(
      -- Reset in
      async_sys_rst            => logic0,
      aclk                     => aclk,
      aresetn                  => aresetn,
      ds_axi_awaddr            => ds_axi_awaddr,
      ds_axi_awlen             => ds_axi_awlen,
      ds_axi_awsize            => ds_axi_awsize,
      ds_axi_awburst           => ds_axi_awburst,
      ds_axi_awcache           => ds_axi_awcache,
      ds_axi_awprot            => ds_axi_awprot,
      ds_axi_awvalid           => ds_axi_awvalid,
      ds_axi_wdata             => ds_axi_wdata,
      ds_axi_wstrb             => ds_axi_wstrb,
      ds_axi_wlast             => ds_axi_wlast,
      ds_axi_wvalid            => ds_axi_wvalid,
      ds_axi_bready            => ds_axi_bready,
      ds_axi_araddr            => ds_axi_araddr,
      ds_axi_arlen             => ds_axi_arlen,
      ds_axi_arsize            => ds_axi_arsize,
      ds_axi_arburst           => ds_axi_arburst,
      ds_axi_arcache           => ds_axi_arcache,
      ds_axi_arprot            => ds_axi_arprot,
      ds_axi_arvalid           => ds_axi_arvalid,
      ds_axi_rready            => ds_axi_rready,
      ds_axi_awready           => ds_axi_awready,
      ds_axi_wready            => ds_axi_wready,
      ds_axi_bresp             => ds_axi_bresp,
      ds_axi_bvalid            => ds_axi_bvalid,
      ds_axi_arready           => ds_axi_arready,
      ds_axi_rdata             => ds_axi_rdata,
      ds_axi_rresp             => ds_axi_rresp,
      ds_axi_rlast             => ds_axi_rlast,
      ds_axi_rvalid            => ds_axi_rvalid,
      ds_axi_awlock            => ds_axi_awlock,
      ds_axi_awqos             => ds_axi_awqos,
      ds_axi_arlock            => ds_axi_arlock,
      ds_axi_arqos             => ds_axi_arqos,
      ds_axi_awregion          => ds_axi_awregion,
      ds_axi_arregion          => ds_axi_arregion,
      mptl_tx_p                => mptl_t2b_p,
      mptl_tx_n                => mptl_t2b_n,
      mptl_rx_p                => mptl_b2t_p,
      mptl_rx_n                => mptl_b2t_n,
      mptl_clk_p               => mptl_clk_p,
      mptl_clk_n               => mptl_clk_n,
      mptl_target_configured_n => mptl_target_configured_l,
      mptl_target_gtp_online_n => mptl_target_gtp_online_l,
      mptl_bridge_gtp_online_n => mptl_bridge_gtp_online_l,
      gpio_t2b                 => mptl_gpio_t2b,
      gpio_b2t                 => mptl_gpio_b2t
    );


  --
  -- Convert from AXI4 128-bit wide to AXI4 32-bit wide.
  --

  axi4_128_to_32_inst : axi4_128_to_32
    port map(
      s_axi_aclk     => aclk,
      s_axi_aresetn  => aresetn,
      s_axi_awaddr   => ds_axi_awaddr,
      s_axi_awlen    => ds_axi_awlen,
      s_axi_awsize   => ds_axi_awsize,
      s_axi_awburst  => ds_axi_awburst,
      s_axi_awlock   => ds_axi_awlock,
      s_axi_awcache  => ds_axi_awcache,
      s_axi_awprot   => ds_axi_awprot,
      s_axi_awregion => ds_axi_awregion,
      s_axi_awqos    => ds_axi_awqos,
      s_axi_awvalid  => ds_axi_awvalid,
      s_axi_awready  => ds_axi_awready,
      s_axi_wdata    => ds_axi_wdata,
      s_axi_wstrb    => ds_axi_wstrb,
      s_axi_wlast    => ds_axi_wlast,
      s_axi_wvalid   => ds_axi_wvalid,
      s_axi_wready   => ds_axi_wready,
      s_axi_bresp    => ds_axi_bresp,
      s_axi_bvalid   => ds_axi_bvalid,
      s_axi_bready   => ds_axi_bready,
      s_axi_araddr   => ds_axi_araddr,
      s_axi_arlen    => ds_axi_arlen,
      s_axi_arsize   => ds_axi_arsize,
      s_axi_arburst  => ds_axi_arburst,
      s_axi_arlock   => ds_axi_arlock,
      s_axi_arcache  => ds_axi_arcache,
      s_axi_arprot   => ds_axi_arprot,
      s_axi_arregion => ds_axi_arregion,
      s_axi_arqos    => ds_axi_arqos,
      s_axi_arvalid  => ds_axi_arvalid,
      s_axi_arready  => ds_axi_arready,
      s_axi_rdata    => ds_axi_rdata,
      s_axi_rresp    => ds_axi_rresp,
      s_axi_rlast    => ds_axi_rlast,
      s_axi_rvalid   => ds_axi_rvalid,
      s_axi_rready   => ds_axi_rready,
      m_axi_awaddr   => ds_axi32_awaddr,
      m_axi_awlen    => ds_axi32_awlen,
      m_axi_awsize   => ds_axi32_awsize,
      m_axi_awburst  => ds_axi32_awburst,
      m_axi_awlock   => ds_axi32_awlock,
      m_axi_awcache  => ds_axi32_awcache,
      m_axi_awprot   => ds_axi32_awprot,
      m_axi_awregion => ds_axi32_awregion,
      m_axi_awqos    => ds_axi32_awqos,
      m_axi_awvalid  => ds_axi32_awvalid,
      m_axi_awready  => ds_axi32_awready,
      m_axi_wdata    => ds_axi32_wdata,
      m_axi_wstrb    => ds_axi32_wstrb,
      m_axi_wlast    => ds_axi32_wlast,
      m_axi_wvalid   => ds_axi32_wvalid,
      m_axi_wready   => ds_axi32_wready,
      m_axi_bresp    => ds_axi32_bresp,
      m_axi_bvalid   => ds_axi32_bvalid,
      m_axi_bready   => ds_axi32_bready,
      m_axi_araddr   => ds_axi32_araddr,
      m_axi_arlen    => ds_axi32_arlen,
      m_axi_arsize   => ds_axi32_arsize,
      m_axi_arburst  => ds_axi32_arburst,
      m_axi_arlock   => ds_axi32_arlock,
      m_axi_arcache  => ds_axi32_arcache,
      m_axi_arprot   => ds_axi32_arprot,
      m_axi_arregion => ds_axi32_arregion,
      m_axi_arqos    => ds_axi32_arqos,
      m_axi_arvalid  => ds_axi32_arvalid,
      m_axi_arready  => ds_axi32_arready,
      m_axi_rdata    => ds_axi32_rdata,
      m_axi_rresp    => ds_axi32_rresp,
      m_axi_rlast    => ds_axi32_rlast,
      m_axi_rvalid   => ds_axi32_rvalid,
      m_axi_rready   => ds_axi32_rready
    );


  --
  -- Convert from AXI4 32-bit wide to AXI4 Lite 32-bit wide.
  --

  axi4_to_axi4l_inst : axi4_to_axi4l
    port map(
      aclk           => aclk,
      aresetn        => aresetn,
      s_axi_awaddr   => ds_axi32_awaddr,
      s_axi_awlen    => ds_axi32_awlen,
      s_axi_awsize   => ds_axi32_awsize,
      s_axi_awburst  => ds_axi32_awburst,
      s_axi_awlock   => ds_axi32_awlock,
      s_axi_awcache  => ds_axi32_awcache,
      s_axi_awprot   => ds_axi32_awprot,
      s_axi_awregion => ds_axi32_awregion,
      s_axi_awqos    => ds_axi32_awqos,
      s_axi_awvalid  => ds_axi32_awvalid,
      s_axi_awready  => ds_axi32_awready,
      s_axi_wdata    => ds_axi32_wdata,
      s_axi_wstrb    => ds_axi32_wstrb,
      s_axi_wlast    => ds_axi32_wlast,
      s_axi_wvalid   => ds_axi32_wvalid,
      s_axi_wready   => ds_axi32_wready,
      s_axi_bresp    => ds_axi32_bresp,
      s_axi_bvalid   => ds_axi32_bvalid,
      s_axi_bready   => ds_axi32_bready,
      s_axi_araddr   => ds_axi32_araddr,
      s_axi_arlen    => ds_axi32_arlen,
      s_axi_arsize   => ds_axi32_arsize,
      s_axi_arburst  => ds_axi32_arburst,
      s_axi_arlock   => ds_axi32_arlock,
      s_axi_arcache  => ds_axi32_arcache,
      s_axi_arprot   => ds_axi32_arprot,
      s_axi_arregion => ds_axi32_arregion,
      s_axi_arqos    => ds_axi32_arqos,
      s_axi_arvalid  => ds_axi32_arvalid,
      s_axi_arready  => ds_axi32_arready,
      s_axi_rdata    => ds_axi32_rdata,
      s_axi_rresp    => ds_axi32_rresp,
      s_axi_rlast    => ds_axi32_rlast,
      s_axi_rvalid   => ds_axi32_rvalid,
      s_axi_rready   => ds_axi32_rready,
      m_axi_awaddr   => ds_axil_awaddr,
      m_axi_awprot   => ds_axil_awprot,
      m_axi_awvalid  => ds_axil_awvalid,
      m_axi_awready  => ds_axil_awready,
      m_axi_wdata    => ds_axil_wdata,
      m_axi_wstrb    => ds_axil_wstrb,
      m_axi_wvalid   => ds_axil_wvalid,
      m_axi_wready   => ds_axil_wready,
      m_axi_bresp    => ds_axil_bresp,
      m_axi_bvalid   => ds_axil_bvalid,
      m_axi_bready   => ds_axil_bready,
      m_axi_araddr   => ds_axil_araddr,
      m_axi_arprot   => ds_axil_arprot,
      m_axi_arvalid  => ds_axil_arvalid,
      m_axi_arready  => ds_axil_arready,
      m_axi_rdata    => ds_axil_rdata,
      m_axi_rresp    => ds_axil_rresp,
      m_axi_rvalid   => ds_axil_rvalid,
      m_axi_rready   => ds_axil_rready
    );


  --
  -- Convert AXI4-Lite 32-bit wide to BRAM-style register signals.
  --
  
  axi4l_to_regs_inst : axi4l_to_regs
    port map(
      s_axi_aclk    => aclk,
      s_axi_aresetn => aresetn,
      s_axi_awaddr  => ds_axil_awaddr(11 downto 0),
      s_axi_awprot  => ds_axil_awprot,
      s_axi_awvalid => ds_axil_awvalid,
      s_axi_awready => ds_axil_awready,
      s_axi_wdata   => ds_axil_wdata,
      s_axi_wstrb   => ds_axil_wstrb,
      s_axi_wvalid  => ds_axil_wvalid,
      s_axi_wready  => ds_axil_wready,
      s_axi_bresp   => ds_axil_bresp,
      s_axi_bvalid  => ds_axil_bvalid,
      s_axi_bready  => ds_axil_bready,
      s_axi_araddr  => ds_axil_araddr(11 downto 0),
      s_axi_arprot  => ds_axil_arprot,
      s_axi_arvalid => ds_axil_arvalid,
      s_axi_arready => ds_axil_arready,
      s_axi_rdata   => ds_axil_rdata,
      s_axi_rresp   => ds_axil_rresp,
      s_axi_rvalid  => ds_axil_rvalid,
      s_axi_rready  => ds_axil_rready,
      bram_rst_a    => reg_rst_a,
      bram_clk_a    => reg_clk_a,
      bram_en_a     => reg_en_a,
      bram_we_a     => reg_we_a,
      bram_addr_a   => reg_addr_a,
      bram_wrdata_a => reg_wrdata_a,
      bram_rddata_a => reg_rddata_a
    );


  --
  -- Implement registers.
  --

  write_regs : process(reg_rst_a, reg_clk_a)
  
    variable istatus_clear : std_logic_vector(REG_DATA_WIDTH-1 downto 0);
  
  begin
  
    if reg_rst_a = '1' then
    
      reg_ienable <= (others => '0');
      
      reg_istatus <= (others => '0');
      
      reg_test <= (others => '0');
      
      reg_period <= (others => '0');
      
      rearm_firq <= '0';
      
    elsif rising_edge(reg_clk_a) then
    
      if reg_addr_a(5 downto 2) = "0000" then
        -- Write to IENABLE
        for i in 0 to REG_BE_WIDTH-1 loop
          if reg_we_a(i) = '1' then
            reg_ienable(8*(i+1)-1 downto 8*i) <= reg_wrdata_a(8*(i+1)-1 downto 8*i);
          end if;
        end loop;
      end if;
      
      if reg_addr_a(5 downto 2) = "0001" then
        -- Write to IENABLE_SET
        for i in 0 to REG_BE_WIDTH-1 loop
          if reg_we_a(i) = '1' then
            reg_ienable(8*(i+1)-1 downto 8*i) <= reg_ienable(8*(i+1)-1 downto 8*i) or reg_wrdata_a(8*(i+1)-1 downto 8*i);
          end if;
        end loop;
      end if;
      
      if reg_addr_a(5 downto 2) = "0010" then
        -- Write to IENABLE_CLR
        for i in 0 to REG_BE_WIDTH-1 loop
          if reg_we_a(i) = '1' then
            reg_ienable(8*(i+1)-1 downto 8*i) <= reg_ienable(8*(i+1)-1 downto 8*i) and not reg_wrdata_a(8*(i+1)-1 downto 8*i);
          end if;
        end loop;
      end if;
      
      istatus_clear := (others => '0');
      if reg_addr_a(5 downto 2) = "0011" then
        -- Write to ISTATUS
        for i in 0 to REG_BE_WIDTH-1 loop
          if reg_we_a(i) = '1' then
            istatus_clear(8*(i+1)-1 downto 8*i) := reg_wrdata_a(8*(i+1)-1 downto 8*i);
          end if;
        end loop;
      end if;
      
      -- If an internal_interrupt is asserted at the same time as somebody tries to
      -- clear it, the corresponding bit in the ISTATUS register remains asserted.      
      reg_istatus <= (reg_istatus and not istatus_clear) or internal_interrupts;
      
      -- Any time somebody writes to the ISTATUS register (to clear interrupts), force
      -- firq_l deasserted for a clock cycle. This ensures that if there are other
      -- internal interrupts still pending, another falling edge will happen on firq_l.
      if reg_addr_a(5 downto 2) = "0011" and OR_reduce(reg_we_a) = '1' then
        rearm_firq <= '1';
      else
        rearm_firq <= '0';
      end if;

      reg_test <= (others => '0');
      if reg_addr_a(5 downto 2) = "0100" then
        -- Write to TEST
        for i in 0 to REG_BE_WIDTH-1 loop
          if reg_we_a(i) = '1' then
            reg_test(8*(i+1)-1 downto 8*i) <= reg_wrdata_a(8*(i+1)-1 downto 8*i);
          end if;
        end loop;
      end if;
      
      if reg_addr_a(5 downto 2) = "1000" then
        -- Write to PERIOD
        for i in 0 to REG_BE_WIDTH-1 loop
          if reg_we_a(i) = '1' then
            reg_period(8*(i+1)-1 downto 8*i) <= reg_wrdata_a(8*(i+1)-1 downto 8*i);
          end if;
        end loop;
      end if;
    end if;
  end process;

  -- Read latency is 1 cycle
  read_regs : process(reg_rst_a, reg_clk_a)
  begin
    if reg_rst_a = '1' then
      reg_rddata_a <= (others => '0');
    elsif rising_edge(reg_clk_a) then
      case reg_addr_a(5 downto 2) is
        when "0000" | "0001" | "0010" =>
          reg_rddata_a <= reg_ienable;
          
        when "0011" =>
          reg_rddata_a <= reg_istatus;
          
        when "1000" =>
          reg_rddata_a <= reg_period;
        
        when others =>
          reg_rddata_a <= (others => '0');
      end case;
    end if;
  end process;


  --
  -- Timer controlled by PERIOD register, with millisecond resolution
  --
  
  -- Generate a pulse every millisecond, assuming 300 MHz clock.
  -- Process in refclk300 clock domain
  gen_ms_pulse : process(refclk300)
    constant MS_CLOCKS : natural := 300000; -- Must be >= 2 for this counter to work correctly
  begin
    if rising_edge(refclk300) then
      if ms_counter_tc = '1' then
        ms_counter <= (others => '0');
        ms_counter_tc <= '0';
      else
        ms_counter <= ms_counter + 1;
        if ms_counter = MS_CLOCKS - 2 then
          ms_counter_tc <= '1';
        else
          ms_counter_tc <= '0';
        end if;
      end if;
    end if;
  end process;
  
  -- Latch ms_counter_tc until cleared by acknowledgement from reg_clk_a clock domain.
  -- Process in refclk300 clock domain
  latch_ms_counter_tc : process(ms_pulse_xfered, refclk300)
  begin
    if ms_pulse_xfered = '1' then -- Asynchronously clear (because signal generated in reg_clk_a clock domain)
      ms_counter_tc_latched <= '0';
    elsif rising_edge(refclk300) then
      -- Set to 1 by ms_counter_tc pulse and stays at 1 until cleared by ms_pulse_xfered
      if ms_counter_tc = '1' then
        ms_counter_tc_latched <= '1';
      end if;
    end if;
  end process;

  -- Transfer millisecond pulse into reg_clk_a clock domain.
  -- Process in reg_clk_a clock domain
  transfer_ms_pulse : process(reg_clk_a)
  begin
    if rising_edge(reg_clk_a) then
      -- Synchronizing stages
      if ms_pulse_xfered = '1' then
        ms_pulse_xfer0 <= '0';
        ms_pulse_xfer1 <= '0';
        ms_pulse_xfer2 <= '0';
        ms_pulse_xfered <= '0';
      else
        ms_pulse_xfer0 <= ms_counter_tc_latched;
        ms_pulse_xfer1 <= ms_pulse_xfer0;
        ms_pulse_xfer2 <= ms_pulse_xfer1;
        ms_pulse_xfered <= ms_pulse_xfer2;
      end if;
    end if;
  end process;
  
  -- Generate a single-cycle pulse every reg_period[30:0]+1 milliseconds, as long as reg_period[31] is 1.
  -- Process in reg_clk_a clock domain
  gen_timer_interrupt : process(reg_rst_a, reg_clk_a)
  begin
    if reg_rst_a = '1' then
      timer_count <= (others => '0');
      timer_interrupt <= '0';
    elsif rising_edge(reg_clk_a) then
      if reg_period(31) = '1' then
        if ms_pulse_xfered = '1' then -- Millisecond pulse
          if timer_count = reg_period(30 downto 0) then
            timer_count <= (others => '0');
          else
            timer_count <= timer_count + 1;
          end if;
        end if;        
      else
        timer_count(30 downto 0) <= (others => '0');
      end if;
      
      if reg_period(31) = '1' and ms_pulse_xfered = '1' and timer_count = reg_period(30 downto 0) then
        timer_interrupt <= '1';
      else
        timer_interrupt <= '0';
      end if;
    end if;
  end process;
  
  
  --
  -- Generate internal interrupts
  --
  
  -- Internal interrupt source 0 is bit 0 of TEST register
  internal_interrupts(0) <= reg_test(0);
  
  -- Internal interrupt source 1 is terminal count signal of timer controlled by PERIOD register
  internal_interrupts(1) <= timer_interrupt;
  
  -- Internal interrupt sources 2 to 31 are unused at present.
  internal_interrupts(31 downto 2) <= (others => '0');
  

  --
  -- To generate the FPGA to host interrupt request, we OR together all enabled
  -- and pending internal interrupts and negate the result.
  --
  -- HOWEVER, we force firq_l deasserted for a clock cycle any time somebody
  -- writes to the ISTATUS register (to clear internal interrupts). This ensures
  -- that if there are still pending internal interrupts, another falling edge
  -- will then occur in firq_l, to alert the host.
  --
  
  generate_firq : process(reg_clk_a)
  begin
    if rising_edge(reg_clk_a) then
      firq_l <= not OR_reduce(reg_ienable and reg_istatus) or -- OR together pending and enabled internal interrupts, and negate
                rearm_firq;                                   -- Force firq_l high on write to ISTATUS
    end if;
  end process;
  
  
  --
  -- Buffer the 300 MHz reference clock
  --
  
  refclk300_ibuf_inst : IBUFGDS
    port map(
      I => refclk300_p,
      IB => refclk300_n,
      O => refclk300_ibuf);
      
  refclk300_bufg_inst : BUFG
    port map(
      I => refclk300_ibuf,
      O => refclk300);
      
      
end architecture;
